'''
JSON token scanner
'''

import re

__all__ = ['make_scanner']


'''
re.DOTALL: Make the '.' special character match any character at all,
including a newline; without this flag, '.' will match anything except a newline.

re.re.VERBOSE: This flag allows you to write regular expressions that look nicer and are more readable
like:
a = re.compile(r"""\d +  # the integral part
                   \.    # the decimal point
                   \d *  # some fractional digits""", re.X)
b = re.compile(r"\d+\.\d*")

and re.MULTILINE:
When specified, the pattern character '^' matches at the beginning
of the string and at the beginning of each line (immediately following each newline);
 and the pattern character '$' matches at the end of the string and at the end of each
 line (immediately preceding each newline)


 (?:pattern) doesn't capture match
'''
NUMBER_RE = re.compile(r'(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?',
                       (re.VERBOSE | re.MULTILINE | re.DOTALL))


def py_make_scanner(context):
    parse_object = context.parse_object
    parse_array = context.parse_array
    parse_string = context.parse_string
    match_number = NUMBER_RE.match
    parse_float = context.parse_float
    parse_int = context.parse_int

    # parse_constant = context.parse_constant

    def _scan_once(string, idx):
        try:
            nextchar = string[idx]
        except IndexError:
            raise StopIteration

        # match string
        if nextchar == '"':
            return parse_string(string, idx + 1)
        # match object
        elif nextchar == '{':
            return parse_object((string, idx + 1), _scan_once)
        # match array
        elif nextchar == '[':
            return parse_array((string, idx + 1), _scan_once)
        # match null
        elif nextchar == 'n' and string[idx:idx + 4] == 'null':
            return None, idx + 4
        # match true
        elif nextchar == 't' and string[idx:idx + 4] == 'true':
            return True, idx + 4
        # match false
        elif nextchar == 'f' and string[idx:idx + 5] == 'false':
            return False, idx + 5

        # regex.match(string[, pos[, endpos]])
        # match the string from the pos position to endposition
        # mactch number
        m = match_number(string, idx)
        if m is not None:
            integer, frac, exp = m.groups()
            # mactch float
            if frac or exp:
                res = parse_float(integer + (frac or '') + (exp or ''))
            # match integer
            else:
                res = parse_int(integer)
            # m.end() is the nextchar pos of end number
            return res, m.end()
        else:
            raise StopIteration
        '''
        elif nextchar == 'N' and string[idx:idx + 3] == 'NaN':
            return parse_constant('NaN'), idx + 3
        elif nextchar == 'I' and string[idx:idx + 9] == 'Infinity':
            return parse_constant('Infinity'), idx + 9
        '''

    return _scan_once


make_scanner = py_make_scanner



from distutils.core import setup
# To use a consistent encoding

setup(
    name='yson',

    version='0.1',
    packages=['yson'],

    # The project's main homepage.
    url='https://github.com/yydai/yson',
    author='yydai',
    author_email='dai92817@icloud.com',

    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 2.7'
    ],
    keywords=['json', 'parser', 'demo']
)



'''
Implementation of JSONDecoder
'''

import re
from yson import scanner

__all__ = ['JSONDecoder']

FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL


def linecol(doc, pos):
    # get line number of current position
    lineno = doc.count('\n', 0, pos) + 1
    if lineno == 1:
        colno = pos + 1
    else:
        # get the column number of current position in current line
        colno = pos - doc.rindex('\n', 0, pos)

    return lineno, colno


WHITESPACE = re.compile(r'[ \t\n\r]*', FLAGS)
WHITESPACE_STR = ' \t\n\r'


# may have bugs
def pass_whitespace(s, idx):
    if (idx < len(s)) and (s[idx] in WHITESPACE_STR):
        idx += 1
        try:
            if s[idx] in WHITESPACE_STR:
                idx = WHITESPACE.match(s, idx).end()
        except IndexError:
            pass
    return idx


BACKSLASH = {
    '"': u'"', '\\': u'\\', '/': u'/',
    'b': u'\b', 'f': u'\f', 'n': u'\n', 'r': u'\r', 't': u'\t',
}

DEFAULT_ENCODING = 'utf-8'


def errmsg(msg, doc, pos, end=None):
    lineno, colno = linecol(doc, pos)
    if end is None:
        fmt = '{0}: line {1} column {2} (char {3})'
        return fmt.format(msg, lineno, colno, pos)
    endlineno, endcolno = linecol(doc, end)
    fmt = '{0}: line {1} column {2} - line {3} column {4} (char {5} - {6})'
    return fmt.format(msg, lineno, colno, endlineno, endcolno, pos, end)


def _decode_uXXXX(s, pos):
    esc = s[pos + 1:pos + 5]
    if len(esc) == 4 and esc[1] not in 'xX':
        try:
            # the 16 is base
            return int(esc, 16)
        except ValueError:
            pass

    msg = "Invalid \\uXXXX escape"
    raise ValueError(errmsg(msg))


def py_scanstring(s, idx, _b=BACKSLASH):
    # [\x00-\x1f] is Ascii control characters, below also include '"' and '\'
    # any unicode character except " or \ or control character
    STRINGCHUNK = re.compile(r'(.*?)(["\\\x00-\x1f])', FLAGS)

    encoding = DEFAULT_ENCODING
    chunks = []
    _append = chunks.append
    begin = idx - 1

    while True:
        chunk = STRINGCHUNK.match(s, idx)
        if chunk is None:
            raise ValueError(
                errmsg('Unterminated string starting at', s, begin))

        idx = chunk.end()
        content, terminator = chunk.groups()
        if content:
            if not isinstance(content, unicode):
                content = unicode(content, encoding)
            _append(content)

        if terminator == '"':
            break

    return ''.join(chunks), idx


def JSONObject(s_and_idx, scan_once, memo=None, w=WHITESPACE.match, _ws=WHITESPACE_STR):
    s, idx = s_and_idx
    if memo is None:
        memo = {}
    # setdefault
    # If key is in the dictionary, return its value. If not,
    # insert key with a value of default and return default.
    # default defaults to None.

    while True:
        idx = pass_whitespace(s, idx)
        nextchar = s[idx]
        if nextchar == '"':
            res, idx = scan_once(s, idx)
        idx = pass_whitespace(s, idx)
        nextchar = s[idx]
        if nextchar != ':':
            raise ValueError(errmsg("Need ':' delimiter"))
        else:
            idx += 1
        idx = pass_whitespace(s, idx)
        try:
            value, idx = scan_once(s, idx)
        except StopIteration:
            raise ValueError(errmsg("Expecting object", s, idx))

        memo[res] = value
        idx = pass_whitespace(s, idx)
        nextchar = s[idx]

        idx += 1
        if nextchar == '}':
            break
        if nextchar != ',':
            raise ValueError(errmsg("Expecting ',' delimiter", s, idx - 1))

        idx = pass_whitespace(s, idx)
    return memo, idx


def JSONArray(s_and_idx, scan_once, _w=WHITESPACE.match, _ws=WHITESPACE_STR):
    s, idx = s_and_idx
    values = []
    idx = pass_whitespace(s, idx)
    nextchar = s[idx]
    if nextchar == ']':
        return values, idx + 1

    _append = values.append

    while True:
        try:
            value, idx = scan_once(s, idx)
        except StopIteration:
            raise ValueError(errmsg("Expecting object", s, idx))
        _append(value)
        idx = pass_whitespace(s, idx)
        nextchar = s[idx:idx + 1]

        idx += 1
        if nextchar == ']':
            break
        if nextchar != ',':
            raise ValueError(errmsg("Expecting ',' delimiter", s, idx - 1))

        idx = pass_whitespace(s, idx)

    return values, idx


class JSONDecoder(object):
    '''
    Performs the following translations in decoding by default:

    +---------------+-------------------+
    | JSON          | Python            |
    +===============+===================+
    | object        | dict              |
    +---------------+-------------------+
    | array         | list              |
    +---------------+-------------------+
    | string        | unicode           |
    +---------------+-------------------+
    | number (int)  | int, long         |
    +---------------+-------------------+
    | number (real) | float             |
    +---------------+-------------------+
    | true          | True              |
    +---------------+-------------------+
    | false         | False             |
    +---------------+-------------------+
    | null          | None              |
    +---------------+-------------------+

    '''

    def __init__(self):
        self.parse_int = int
        self.parse_float = float
        self.parse_object = JSONObject
        self.parse_array = JSONArray
        self.parse_string = py_scanstring
        self.scan_once = scanner.make_scanner(self)

    def decode(self, s, _w=WHITESPACE.match):
        # start from nonspace char
        obj, end = self.raw_decode(s, idx=_w(s, 0).end())
        end = _w(s, end).end()
        if end != len(s):
            raise ValueError(errmsg("Extra data", s, end, len(s)))
        return obj

    def raw_decode(self, s, idx=0):
        try:
            obj, end = self.scan_once(s, idx)
        except StopIteration:
            raise ValueError("No JSON object could be decoded")
        return obj, end


if __name__ == '__main__':
    j = JSONDecoder()
    a = j.decode('[12, 23, 45.3, "hello world", {"aaa": 123}]')
    b = '''{
    "name": "dai ying",
    "age" : 20,
    "family": {
    "father": "Dai Hongjun",
    "mother": "zhenbenying"
    },
    "hobby": ["game", "guitar", "Emacs"]}
    '''
    b = j.decode(b)
    print type(a)
    print a
    print type(b)
    print b
